from bottlenest.core.NestModuleContext import NestModuleContext
from bottlenest.core.NestLogger import NestLogger
from bottlenest.transports.http.HttpTransport import HttpTransport


class NestApplicationContext:
    def __init__(self, module, transport):
        self.logger = self.__setupLogger()
        self.module = module
        self.module.enableModule()
        self.moduleName = module.moduleName
        self.moduleContext = self.__setupModuleContext(
            module=module,
            logger=self.logger,
        )
        # Calls NestHttpModule.setup()
        # self.setup()
        self.module.setupModule(
            appContext=self,
            moduleContext=self.moduleContext,
            transport=transport,
        )

    # def setup(self):
    #     # self.app = self.transport
    #     # self.moduleContext.set('transport', self.transport)
    #     # self.moduleContext.set('NestApplicationContext', self)
    #     # self.moduleContext.set('NestModuleContext', self.moduleContext)
    #     # Calls NestHttpModule.setup()
    #     self.module.setup(self.moduleContext)

    def __setupLogger(self):
        logger = NestLogger()
        return logger

    def __setupModuleContext(self, module, logger):
        moduleContext = NestModuleContext()
        moduleContext.set('module', module)
        moduleContext.set('logger', logger)
        return moduleContext

    def listen(self):
        self.logger.log('NestApplicationContext listen')
        self.module.listen()



from bottlenest.common import Controller, Get
from .AppService import AppService


@Controller()
class AppController:
    def __init__(self, context):
        self.appService = context.inject(AppService)

    @Get('/')
    def getHello(self, req):
        return self.appService.getHello()

    @Get('/hello/:name')
    def getHello2(self, req):
        return f"Hello {req.params.name}!"



from flask import Flask
from .errors import HttpError
import traceback
import eventlet
from eventlet import wsgi


class HttpTransport:
    def __init__(self, port=3500):
        self.port = port
        self.app = None

    # called automatically
    # when you run NestFactory.createMicroservice
    # (inside NestApplicationContext)
    def setupTransport(self, appContext, moduleContext):
        print("HttpTransport setupTransport")
        self.appContext = appContext
        self.moduleContext = moduleContext
        self.logger = self.moduleContext.get('logger')
        self.app = Flask(self.appContext.module.moduleName)
        # self.appContext.set('app', self.app)
        # if isinstance(self.moduleContext.get('transport'), HttpTransport):
        #     print("---------------------> setting up error handlers")
        #     self.setupErrorHandlers()

    # handle any http errors
    def setupErrorHandlers(self):
        @self.app.errorhandler(Exception)
        def defaultErrorHandler(e):
            self.logger.log(
                'NestApplicationContext handle_exception', e, traceback.format_exc())
            # check if e is an instance of HttpError
            if isinstance(e, HttpError):
                return e.toDict(), e.statusCode
            return {"messages": [e.__str__()]}, 500

    # start listening for requests
    # this is called
    def listen(self, pool, callback):
        print(f"HttpTransport listen port={self.port}")
        # self.app.run(port=self.port, debug=False)
        # spawned = pool.spawn(self.app.run, port=self.port, debug=False)

        def _startHttpServer(port, app):
            wsgi.server(eventlet.listen(('0.0.0.0', port)), app)
        pool.spawn(_startHttpServer, self.port, self.app)
        callback()
        # return self.app.run(port=self.port, debug=False)



import inquirer
from bottlenest.metaClasses.NestProvider import NestProvider
from bottlenest.transports.cli.CommandFactory import CommandFactory
from bottlenest.transports.cli.decorators.NestCommandArgument import NestCommandArgument
import sys


class NestCommand(NestProvider):
    __name__ = 'NestCommand'

    def __init__(self, cls, commandName, description):
        self.cls = cls
        self.commandName = commandName
        self.description = description
        CommandFactory.register(self)

    def eventName(self):
        # return NestRoute.__name__
        return self.commandName

    def getName(self):
        return self.commandName

    def parseArguments(self, parser):
        arguments = [ag for ag in dir(self.cls) if isinstance(
            getattr(self.cls, ag), NestCommandArgument)]
        for argumentName in arguments:
            argument = getattr(self.cls, argumentName)
            print(f"argument: {argumentName} {argument}")
            if argument.optional is False:
                parser.add_argument(
                    argument.argumentName,
                    **argument.kwargs,
                    required=True,
                )
            parser.add_argument(
                argument.argumentName,
                **argument.kwargs,
            )

    def setupProvider(self, module, context):
        self._setupProvider(module, context)
        # args = sys.argv[1:]
        # self.provider.run(inquirer, args)



from bottlenest.common import Module
from .AppService import AppService
from .AppController import AppController


@Module(
    imports=[],
    providers=[AppService],
    controllers=[AppController],
)
class AppModule:
    pass



from bottlenest.core.NestModuleContext import NestModuleContext
from bottlenest.core.NestLogger import NestLogger
import sys
import inquirer
import argparse


class CommandFactory:
    __name__ = 'CommandFactory'

    __commands__ = {}
    # __args__ = None
    # __currentCommand__ = None
    # __args__ = sys.argv[1:]
    # __module__ = None
    # __container__ = None

    @staticmethod
    def run(module):
        """This run will be called from whithin main.py"""
        # initial setup
        logger = NestLogger()
        container = CommandFactory.__buildContainer()
        container.set('module', module)
        container.set('logger', logger)
        container.set('inquirer', inquirer)
        # load module

        def __setupModuleContext(module, logger):
            moduleContext = NestModuleContext()
            moduleContext.set('module', module)
            moduleContext.set('logger', logger)
            moduleContext.module = module  # TODO: move this
            return moduleContext
        moduleContext = __setupModuleContext(
            module=module,
            logger=logger,
        )
        module.enableModule()
        # TODO: review this
        module.setupModule(
            moduleContext,
            container,
        )
        # parse initial command line arguments
        # and set __currentCommand__
        rawCommandName = sys.argv[1]
        # CommandFactory.__module__ = module
        # CommandFactory.__container__ = container
        # run command
        CommandFactory.__runCommand(container, rawCommandName)

    @staticmethod
    def __buildContainer():
        context = NestModuleContext()
        parser = argparse.ArgumentParser(
            prog="Program Name",
            description="Program Description",
            epilog="Program Epilog",
        )
        # TODO:GIOVANNEFEITOSA
        parser.add_argument("command", help="command to run")
        # parser.add_argument("args", nargs=argparse.REMAINDER)
        # augment context
        context.parser = parser
        context.inquirer = inquirer
        return context

    @staticmethod
    def __runCommand(context, commandName):
        """This runCommand will be called from whithin NestCommand"""
        print("CommandFactory runCommand ", commandName)
        # parse command line arguments
        # parser = CommandFactory.__buildParser(context)
        # help command
        if commandName == 'help':
            context.parser.print_help()
            return
        # run command
        command = CommandFactory.__commands__[commandName]
        command.parseArguments(context.parser)
        commandArgs = context.parser.parse_args()
        command.cls.run(command, context, commandArgs)

    @staticmethod
    def register(nestCommand):
        """This register will be called from whithin NestCommand"""
        print("CommandFactory register ", nestCommand.commandName)
        CommandFactory.__commands__[nestCommand.commandName] = nestCommand



from bottlenest.transports.websockets.decorators.NestWebSocketGateway import NestWebSocketGateway
from bottlenest.transports.websockets.decorators.NestSubscribeMessage import NestSubscribeMessage

# música boa
# https://www.youtube.com/watch?v=KCwk1qeh0gM
# https://www.youtube.com/watch?v=Xj4ETM2IJY8&list=PLTeMoz1bBorZjs92TBazC_QSvGOfw-CJS&index=7


def WebSocketGateway(port=80, namespace=None):
    def wrapper(gatewayClass):
        return NestWebSocketGateway(
            cls=gatewayClass,
            port=port,
            namespace=namespace,
        )
    return wrapper


def SubscribeMessage(eventName):
    def wrapper(messageClass):
        return NestSubscribeMessage(
            callback=messageClass,
            eventName=eventName,
        )
    return wrapper



from ..NestHttpModule import NestHttpModule
from .NestController import NestController
from .NestRoute import NestRoute


def Module(controllers=[], imports=[], providers=[]):
    def wrapper(moduleClass):
        return NestHttpModule(
            moduleClass=moduleClass,
            imports=imports,
            controllers=controllers,
            providers=providers,
        )
    return wrapper


def Controller():
    def wrapper(controllerClass):
        return NestController(
            cls=controllerClass,
        )
    return wrapper


def Get(path):
    print(f"get defined {path}")

    def wrapper(func):
        return NestRoute(
            callback=func,
            path=path,
            method='GET',
        )
    return wrapper


def Post(path):
    print(f"post defined {path}")

    def wrapper(func):
        return NestRoute(
            callback=func,
            path=path,
            method='POST',
        )
    return wrapper


def Put(path):
    print(f"put defined {path}")

    def wrapper(func):
        return NestRoute(
            callback=func,
            path=path,
            method='PUT',
        )
    return wrapper


def Delete(path):
    print(f"delete defined {path}")

    def wrapper(func):
        return NestRoute(
            callback=func,
            path=path,
            method='DELETE',
        )
    return wrapper


def Patch(path):
    print(f"patch defined {path}")

    def wrapper(func):
        return NestRoute(
            callback=func,
            path=path,
            method='PATCH',
        )
    return wrapper



from ..metaClasses.NestProvider import NestProvider


def Injectable():
    def wrapper(originalClass):
        return NestInjectable(
            originalClass=originalClass,
        )
    return wrapper


class NestInjectable(NestProvider):
    __name__ = 'NestInjectable'

    def __init__(self, originalClass):
        # self.__name__ = originalClass.__name__
        self.cls = originalClass
        self.providerName = originalClass.__name__
        self.originalClass = originalClass
        self.classInstance = None
        # self.dependencies = []

    def getName(self):
        return self.providerName

    def eventName(self):
        return self.providerName

    # def enableProvider(self):
    #    pass

    def setupInjectable(self, module, container):
        print(
            f"NestInjectable setupInjectable {self.providerName} {container}")
        # self.module = module
        # self.container = container
        # TODO: This should be a singleton manager
        self.classInstance = self.originalClass(container)
        # providerName = f"{module.moduleName}.{self.providerName}"
        providerName = self.providerName
        container.set(providerName, self)

    def __repr__(self):
        return f"{self.providerName}()"

    def __str__(self):
        return f"{self.providerName}()"



class NestModuleContext:
    def __init__(self):
        self.container = {}

    def set(self, key, value):
        print("NestModuleContext set", key, value)
        self.container[key] = value

    def get(self, key):
        print("NestModuleContext get", key)
        return self.container[key] if key in self.container else None

    # def inject(self, injectable):
    #     moduleName = self.container['module'].name
    #     providerName = injectable.__name__
    #     key = f"{moduleName}.{providerName}"
    #     provider = self.container[key] if key in self.container else None
    #     if provider:
    #         return provider.instance
    #     else:
    #         raise Exception(f"Provider not found: {key}")



class NestLogger:
    def __init__(self):
        # self.context = None
        # self.log('NestLogger initialized')
        pass

    # def setContext(self, context):
    #     self.context = context
    #     # self.log('NestLogger context set')

    def log(self, *args):
        print(*args)



from bottlenest.transports.cli.decorators.NestCommand import NestCommand
from bottlenest.transports.cli.decorators.NestCommandArgument import NestCommandArgument


def Command(name, *args, **kwargs):
    def decorator(cls):
        return NestCommand(
            cls=cls,
            commandName=name,
            *args,
            **kwargs,
        )
    return decorator


def CommandArgument(name, optional=None, *args, **kwargs):
    def decorator(callback):
        return NestCommandArgument(
            callback=callback,
            argumentName=name,
            optional=optional,
            *args,
            **kwargs,
        )
    return decorator



class NestSubscribeMessage:
    def __init__(self, callback, eventName):
        self.eventName = eventName
        self.callback = callback

    # called from whithin NestWebsocketGateway._setupProvider()
    def setupEvent(self, cls, context):
        print(f"setup event {self.eventName}")
        sio = context.get('sio')

        @sio.on(self.eventName)
        def _callback(sid, data):
            print(f"NestSubscribeMessage {self.eventName}")
            result = self.callback(cls, data)
            if result is not None:
                sio.emit(self.eventName, result, room=sid)
